# -*- coding: utf-8 -*-

"""The base class for nodes (steps) in flowcharts.
"""

import bibtexparser
import calendar
import collections.abc
import hashlib

try:
    import importlib.metadata as implib
except Exception:
    import importlib_metadata as implib
import jinja2
import json
import logging
import os.path
from pathlib import Path
import pprint
import string
import traceback

import numpy as np
import pandas
import uuid

import reference_handler
import seamm
from seamm_util import Q_
import seamm_util
from seamm_util.printing import FormattedText as __
import seamm_util.printing as printing

logger = logging.getLogger(__name__)
job = printing.getPrinter()


def scale(data, factor):
    """Recursive helper to scale e.g. nested lists by a factor."""
    result = []
    for value in data:
        if isinstance(value, list):
            result.append(scale(value, factor))
        else:
            result.append(value * factor)
    return result


class Node(collections.abc.Hashable):
    """The base class for nodes (steps) in flowcharts.

    Parameters
    __________
    flowchart : seamm.Flowchart, optional
        The Flowchart object that contained this node.
    title : str, optional
        The title of this step for use in output.
    extension : str, optional
        Data used in serializing to the flowchart.
    module : str, optional
        The module for this step.
    logger : logging.Logger, optional
        The logger to use for (debug) output. Defaults to the gloabl logger in the
        module.
    uid : str, optional
        The unigue ID for the step, used when reading a flowchart. If not given it is
        generated using uuid.uuid4().

    calculation
    description
    directory
    extension

    flowchart : seamm.Flowchart
        The flowchart that this step is part of.

    global_options
    header
    indent
    job_path

    logger : logging.Logger
        The logger for debug, etc. output.

    metadata
    method
    model
    options

    parent : seamm.Node
        The node that is the parent, usually because this node is in a subflowchart of
        the parent.

    parameters : seamm.Parameters
        The control parameters for this step.

    references
    step_type
    tables
    title

    x : int
        The x-coordinate of the step in the GUI

    y : int
        The y-coordinate of the step in the GUI

    w : int
        The width of the step in the GUI

    h : int
        The height of the step in the GUI

    uuid
    visited

    Notes
    -----
    Handling results
    ~~~~~~~~~~~~~~~~
    The `Node` class takes most or all of the effort out of handling the results of
    calculations in steps. The developer needs to specify the information about possible
    results in `metadata["results"]`. The keys are the steps internal names of the
    results, which match those in the data passed to `store_results()`. The fields of
    the dict for each key give human readable names, units, dimensions, etc.

    The results can be filtered by the `calculation` attribute, and if further filtering
    is needed, the `method` attribute. While these two attirbutes have suggestive names,
    they are simply tags that should match those in the **calculation** filed of the the
    results metadata.

    The `model` attribute is used to form the property name for the database. Often
    results depend on the model chemistry or something similar. The property names
    consist of up to three parts: the property, such as dipole moment; how the property
    was obtained, which is either experiment or the name of the code fo calculated
    results; and the model used, which is usually the model chemistry for calculated
    results.
    """

    def __init__(
        self,
        flowchart=None,
        title="",
        extension=None,
        module=None,
        logger=logger,
        uid=None,
    ):
        if uid is None:
            uid = uuid.uuid4().int

        self._all_options = {}  # All options
        self._bibliography = {}
        self._description = ""
        self._graphs = None
        self._id = None
        self._jinja_env = None
        self._references = None
        self._step_type = None
        self._tables = []
        self._title = title
        self._uuid = uid
        self._visited = False
        self.extension = extension
        self.flowchart = flowchart
        self.logger = logger
        self.parent = None
        self.citation_level = 2  # The level of citation for this code

        # for handling the properties that are results of calculations
        self._metadata = {}
        self._calculation = None  # A tag used to find with results might be created.
        self._method = None  # An optional method such as 'ccsd' or 'dft'.
        self._model = None  # A tag, like 'PM7' or 'MP2/6-31g*', for the model chemistry

        self.parameters = None  # Object containing control parameters

        # Coordinates in the GUI
        self.x = None
        self.y = None
        self.w = None
        self.h = None

        # Set up our formatter for printing
        self.formatter = logging.Formatter(fmt="{message:s}", style="{")

        # Setup the bibliography
        package = self.__module__.split(".")[0]
        files = [p for p in implib.files(package) if "references.bib" in str(p)]
        if len(files) > 0:
            path = files[0].locate()
            self.logger.info(f"bibliography file path = '{path}'")

            data = path.read_text()
            tmp = bibtexparser.loads(data).entries_dict
            writer = bibtexparser.bwriter.BibTexWriter()
            for key, data in tmp.items():
                self.logger.info(f"      {key}")
                self._bibliography[key] = writer._entry_to_bibtex(data)
            self.logger.debug("Bibliography\n" + pprint.pformat(self._bibliography))

    def __hash__(self):
        """Make iterable!"""
        return self._uuid

    def __eq__(self, other):
        return self.__class__ == other.__class__ and self.digest() == other.digest()

    @property
    def all_options(self):
        """The complete set of all options."""
        return self._all_options

    @all_options.setter
    def all_options(self, value):
        self._all_options = value

    @property
    def calculation(self):
        """The type of calculation for filtering available results."""
        return self._calculation

    @property
    def description(self):
        """A textual description of this node"""
        return self._description

    @description.setter
    def description(self, value):
        self._description = value

    @property
    def directory(self):
        """The directory for output and files for this step."""
        return os.path.join(self.flowchart.root_directory, *self._id)

    @property
    def global_options(self):
        """Dictionary of global options"""
        if "SEAMM" not in self.all_options:
            self.global_options = {}
        return self._all_options["SEAMM"]

    @global_options.setter
    def global_options(self, value):
        self._all_options["SEAMM"] = value

    @property
    def header(self):
        """A printable header for this section of output"""
        return "Step {}: {}  {}".format(
            ".".join(str(e) for e in self._id), self.title, self.version
        )

    @property
    def indent(self):
        """The amount to indent the output of this step in **job.out**."""
        length = len(self._id)
        if length <= 1:
            return ""
        if length > 2:
            result = (length - 2) * (3 * " " + ".") + 4 * " "
        else:
            result = 4 * " "
        return result

    @property
    def job_path(self):
        """Return the path to the job's top-level directory"""
        return Path(self.flowchart.root_directory)

    @property
    def metadata(self):
        """Metadata describing aspects of the calculation.

        The metadata is a dictionary of various types of metdata, often themselves
        dictionaries. Common types of metadata are:

        Parameters
        ----------
        keywords
            The keywords for programs with keyword-based input.
        results
            The results that this step can produce.
        """
        return self._metadata

    @property
    def method(self):
        """The method of a calculation, used for filtering metadata.

        A `calculation`, such as **energy** or **optimization**, might return different
        results depending on the type of calculation or how it is carried out. The
        `method` can be used in the metadata to further filter the calculation results.
        """
        return self._method

    @method.setter
    def method(self, value):
        self._method = value

    @property
    def model(self):
        """The model (chemistry) used to obtain results.

        Properties in the database use a trinomial naming scheme:
            `property`#`code`#`model`

        This is the last part of the name, or None if it is not relevant. It is often
        the `model chemistry`, such as **mp2/6-31g** or **PM7**.
        """
        return self._model

    @model.setter
    def model(self, value):
        self._model = value

    @property
    def options(self):
        """Dictionary of options for this step"""
        step_type = self.step_type
        if step_type not in self._all_options:
            self.options = {}
        return self._all_options[step_type]

    @options.setter
    def options(self, value):
        step_type = self.step_type
        self._all_options[step_type] = value

    @property
    def references(self):
        """The reference handler for citations."""
        if self._references is None:
            filename = os.path.join(self.flowchart.root_directory, "references.db")
            self._references = reference_handler.Reference_Handler(filename)

        return self._references

    @references.setter
    def references(self, value):
        if self._references is not None:
            self._references.conn.commit()

        self._references = value

    @property
    def step_type(self):
        """The step type, e.g. 'lammps-step', used for e.g. options"""
        if self._step_type is None:
            name = self.__module__.split(".")[0].replace("_", "-")
            if name == "seamm":
                name = self.__module__.split(".")[1].replace("_", "-")
                name += "-step"
            self._step_type = name

        return self._step_type

    @property
    def tables(self):
        """Any tables this step creates.

        A list of tables this step creates. If it is not easy to decide whether the
        tables are created or just used here, add them to the list. This data is used
        by subsequent steps to present possible tables in the GUI.
        """
        return self._tables

    @tables.setter
    def tables(self, value):
        self._tables = [v for v in value if v.strip() != ""]

    @property
    def tag(self):
        """The string representation of the uuid of the node"""
        return "node=" + str(self._uuid)

    @property
    def title(self):
        """The title to display"""
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def uuid(self):
        """The uuid of the node to give it a unique id."""
        return self._uuid

    @property
    def visited(self):
        """Whether this node has been visited in a traversal"""
        return self._visited

    @visited.setter
    def visited(self, value):
        self._visited = bool(value)

    @staticmethod
    def is_expr(value):
        """Return whether the value is an expression or constant.

        Parameters
        ----------
        value : str
            The value to test

        Returns
        -------
        bool
           True for an expression, False otherwise.
        """
        return len(value) > 0 and value[0] == "$"

    def set_uuid(self):
        self._uuid = uuid.uuid4().int

        # Need to correct all edges to other nodes
        raise NotImplementedError("set_uuid not implemented yet!")

    def set_id(self, node_id):
        """Set the id for node to a given tuple"""
        if self.visited:
            return None
        else:
            self.visited = True
            self._id = node_id
            return self.next()

    def reset_id(self):
        """Reset the id for node"""
        self._id = None

    def get_gui_data(self, key, gui=None):
        """Return an element from the GUI dictionary"""
        if gui is None:
            return self._gui_data[seamm.Flowchart.graphics][key]
        else:
            return self._gui_data[gui][key]

    def set_gui_data(self, key, value, gui=None):
        """Set an element of the GUI dictionary"""
        if gui is None:
            if seamm.Flowchart.graphics not in self._gui_data:
                self._gui_data[seamm.Flowchart.graphics] = {}
            self._gui_data[seamm.Flowchart.graphics][key] = value
        else:
            if gui not in self._gui_data:
                self._gui_data[gui] = {}
            self._gui_data[gui][key] = value

    def get_system_configuration(
        self,
        P=None,
        same_as="current",
        first=True,
        **kwargs,
    ):
        """Get the current system and configuration.

        Optionally use the standard structure handling to create new
        configuration or new system and configuration based on user
        input.

        Note that if the system or configuration do not exist, they are
        automatically created as needed. This allows flowcharts to be started
        with and empty system database and "do the right" thing if a plug-in
        wants to use the curent configuration.

        Parameters
        ----------
        P : dict(str, any) = None
            The dictionary of options and values. If none, the default system and
            configuration are returned as-is.
        same_as : _Configuration = "current""
            Share atoms, bonds, or cell with this configuration, depending
            on other flags. Defaults to "current", which results in using the current
            configuration. If None, an empty configuration is created/used
        first : bool = True
            First configuration of several, which can have different handling than
            the subsequent ones.

        Returns
        -------
        (System, Configuration)
            The system and configuration.
        """
        # Get the system
        system_db = self.get_variable("_system_db")

        system = system_db.system
        if system is None:
            configuration = None
        else:
            configuration = system.configuration
        if same_as == "current":
            same_as = configuration

        if P is None:
            # Just return the current system and configuration, creating if needed.
            if system is None:
                system = system_db.create_system()
            if configuration is None:
                configuration = system.create_configuration()
        else:
            # Honor the user's request for how to handle the structure.
            # If the system or configuration do not exist they are automatically
            # created.
            if first:
                if "structure handling" in P:
                    handling = P["structure handling"]
                else:
                    handling = "Overwrite the current configuration"
            else:
                if "subsequent structure handling" in P:
                    handling = P["subsequent structure handling"]
                else:
                    handling = "Create a new configuration"

            if handling == "Overwrite the current configuration":
                if system is None:
                    system = system_db.create_system()
                if configuration is None:
                    configuration = system.create_configuration()
            elif handling == "Create a new configuration":
                if system is None:
                    system = system_db.create_system()
                # See if sharing atoms, bonds and/or cell
                if same_as is None:
                    configuration = system.create_configuration()
                else:
                    configuration = system.copy_configuration(
                        configuration=same_as, make_current=True
                    )
            elif handling == "Create a new system and configuration":
                system = system_db.create_system()
                if same_as is None:
                    configuration = system.create_configuration()
                else:
                    configuration = system.copy_configuration(
                        configuration=same_as, make_current=True
                    )
            else:
                raise ValueError(
                    f"Do not understand how to handle the structure: '{handling}'"
                )

            # Attend to naming
            if "system name" in P:
                if P["system name"] == "keep current name":
                    pass
                elif P["system name"] == "use SMILES string":
                    system.name = configuration.smiles
                elif P["system name"] == "use Canonical SMILES string":
                    system.name = configuration.canonical_smiles
                else:
                    # Presume it is a string, perhaps with variables.
                    if len(kwargs) == 0:
                        system.name = str(P["system name"])
                    else:
                        system.name = str(P["system name"]).format(**kwargs)
            if "configuration name" in P:
                if P["configuration name"] == "keep current name":
                    pass
                elif P["configuration name"] == "use SMILES string":
                    configuration.name = configuration.smiles
                elif P["configuration name"] == "use Canonical SMILES string":
                    configuration.name = configuration.canonical_smiles
                else:
                    # Presume it is a string, perhaps with variables.
                    if len(kwargs) == 0:
                        configuration.name = str(P["configuration name"])
                    else:
                        configuration.name = str(P["configuration name"]).format(
                            **kwargs
                        )

        return (system, configuration)

    def get_table(self, tablename, create=True):
        """Get the named table, creating if necessary"""
        if not self.variable_exists(tablename):
            # Create the table if allowed to.
            if not create:
                raise RuntimeError(f"Table {tablename} does not exist.")
            table = pandas.DataFrame()
            self.set_variable(
                tablename,
                {
                    "type": "pandas",
                    "table": table,
                    "defaults": {},
                    "loop index": False,
                    "current index": 0,
                    "index column": None,
                },
            )
        table_handle = self.get_variable(tablename)
        return table_handle["table"]

    def connections(self):
        """Return a list of all the incoming and outgoing edges
        for this node, giving the anchor points and other node
        """

        result = self.flowchart.edges(self)
        return result

    def remove_edge(self, edge):
        """Remove a given edge, or all edges if 'all' is given"""

        if isinstance(edge, str) and edge == "all":
            for direction, obj in self.connections():
                self.remove_edge(obj)
        else:
            self.flowchart.graph.remove_edge(
                edge.node1, edge.node2, edge.edge_type, edge.edge_subtype
            )

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
        do.

        Parameters
        ----------
        P :
            a dictionary of parameter values, which may be variables
            or final values. If None, then the parameters values will
            be used as is.
        """
        return (
            "This node has no specific description. "
            "Override the method 'description_text' "
            "to provide the description."
        )

    def describe(self):
        """Write out information about what this node will do"""

        self.visited = True

        # The description
        job.normal(__(self.description_text(), indent=self.indent))

        next_node = self.next()

        if next_node is None or next_node.visited:
            return None
        else:
            return next_node

    def digest(self, strict=False):
        """Generate a unique hash key for this node.

        Parameters
        ----------
        strict: bool
            Whether to include version information. Default: False

        Returns
        -------
        string
        """
        hasher = hashlib.sha256()
        if strict:
            hasher.update(bytes(self.version, "utf-8"))

        for key in self.__dict__:
            if key == "subflowchart":
                # Have a subflowchart!
                hasher.update(bytes(self.__dict__[key].digest(strict=strict), "utf-8"))
            elif key == "parameters":
                if self.parameters is not None:
                    hasher.update(bytes(str(self.parameters.to_dict()), "utf-8"))
                else:
                    if self.__class__.__name__ not in (
                        "StartNode",
                        "LAMMPS",
                        "MOPAC",
                        "Psi4",
                        "Join",
                        "Table",
                    ):
                        print(f"{self.__class__.__name__} has no parameters")

        return hasher.hexdigest()

    def existing_tables(self):
        """Tables from previous steps in the flowchart.

        Returns
        -------
        [str]
            Sorted list of existing tables.
        """
        tables = set()
        if self.parent is not None:
            tables.update(self.parent.existing_tables())

        for node in self.flowchart.get_nodes():
            if node == self:
                break
            tables.update(node.tables)

        return sorted(tables)

    def run(self, printer=None):
        """Do whatever we need to do! The base class does nothing except
        return the next node.
        """

        # Create the directory for writing output and files
        os.makedirs(self.directory, exist_ok=True)

        if printer is not None:
            # Setup up the printing for this step
            self.setup_printing(printer)

        # Add a citation for this plug-in
        package = self.__module__.split(".")[0]
        if package in self._bibliography:
            try:
                template = string.Template(self._bibliography[package])

                version = self.version
                year, month = version.split(".")[0:2]
                month = calendar.month_abbr[int(month)].lower()
                citation = template.substitute(month=month, version=version, year=year)

                title = package.split("_")
                title = " ".join([s.capitalize() for s in title[0:-2]])
                self.references.cite(
                    raw=citation,
                    alias=package,
                    module=package,
                    level=self.citation_level,
                    note=(f"The principle citation for the {title} step in " "SEAMM."),
                )

            except Exception as e:
                printer.important(f"Exception in citation {type(e)}: {e}")
                printer.important(traceback.format_exc())

        next_node = self.next()
        if next_node:
            self.logger.debug("returning next_node: {}".format(next_node))
        else:
            self.logger.debug("returning next_node: None")

        self.close_printing(printer)

        return next_node

    def next(self):
        """Return the next node in the flow"""

        for edge in self.flowchart.edges(self, direction="out"):
            if edge.edge_subtype == "next":
                self.logger.debug("Next node is: {}".format(edge.node2))
                return edge.node2

        self.logger.debug("Reached the end of the flowchart")
        return None

    def previous(self):
        """Return the previous node in the flow"""

        for edge in self.flowchart.edges(self, direction="in"):
            if edge.edge_type == "execution" and edge.edge_subtype == "next":
                return edge.node1

        return None

    def get_input(self):
        """Return the input from this subnode, usually used for
        building up the input for the executable."""

        return ""

    def to_json(self):
        return json.dumps(self.to_dict(), cls=seamm_util.JSONEncoder)

    def to_dict(self):
        """serialize this object and everything it contains as a dict"""
        data = {
            "item": "object",
            "module": self.__module__,
            "class": self.__class__.__name__,
            "version": self.version,
            "extension": self.extension,
        }
        data["attributes"] = {}
        for key in self.__dict__:
            # Remove unneeded variables
            if key[0] == "_" and key not in ("_uuid", "_method", "_tables", "_title"):
                # _method needed because forcefield_step/forcefield.py does not
                # use parameters yet!
                continue
            if key in (
                "bibliography",
                "flowchart",
                "formatter",
                "logger",
                "options",
                "parent",
                "parser",
                "tmp_table",
                "unknown",
            ):
                continue

            if "flowchart" in key:
                # Have a subflowchart!
                data[key] = self.__dict__[key].to_dict()
            else:
                data["attributes"][key] = self.__dict__[key]
        return data

    def from_dict(self, data):
        """un-serialize object and everything it contains from a dict"""
        if data["item"] != "object":
            raise RuntimeError("The data for restoring the object is invalid")
        if data["class"] != self.__class__.__name__:
            raise RuntimeError(
                "Trying to restore a {}".format(self.__class__.__name__)
                + " from data for a {}".format(data["class"])
            )
        for key in data:
            if key == "attributes":
                attributes = data["attributes"]
                for subkey in attributes:
                    self.__dict__[subkey] = attributes[subkey]
            elif "flowchart" in key:
                self.__dict__[key].from_dict(data[key])

    def default_edge_subtype(self):
        """Return the default subtype of the edge. Usually this is 'next'
        but for nodes with two or more edges leaving them, such as a loop, this
        method will return an appropriate default for the current edge. For
        example, by default the first edge emanating from a loop-node is the
        'loop' edge; the second, the 'exit' edge.

        A return value of 'too many' indicates that the node exceeds the number
        of allowed exit edges.
        """

        # how many outgoing edges are there?
        n_edges = len(self.flowchart.edges(self, direction="out"))

        self.logger.debug("node.default_edge_subtype, n_edges = {}".format(n_edges))

        if n_edges == 0:
            return ""
        else:
            return "too many"

    def analyze(self, indent="", **kwargs):
        """Analyze the output of the calculation"""
        return

    def file_path(self, filename):
        """Remove any prefix from a filename and return the path.

        Parameters
        ----------
        filename : str
            The filename with optional prefix such as 'job:'

        Returns
        -------
        pathlib.Path
            The normalized, full path.
        """
        path = str(filename)
        if path[0:4] == "job:":
            path = path[4:]
            path = self.job_path / path
        else:
            path = Path(filename)

        path = path.expanduser().resolve()

        return path

    def get_value(self, variable_or_value):
        """Return the value of the workspace variable is <variable_or_value>
        is the name of a variable. Otherwise, simply return the value of
        <variable_or_value>.

        This provides a convenient way to handle both values and variables
        in widgets. A reference to a variable is $<name> or ${name}, and is
        replaced by the contents of the variable. If the text is not a
        reference to a variable then the value passed in is returned
        unchanged.
        """

        return seamm.flowchart_variables.value(variable_or_value)

    def get_variable(self, variable):
        """Get the value of a variable, which must exist"""

        return seamm.flowchart_variables.get_variable(variable)

    def set_variable(self, variable, value):
        """Set the value of a variable in the workspace. The name of the
        variable maybe a plain string, or be $<name> or ${<name>}
        """

        seamm.flowchart_variables.set_variable(variable, value)

    def variable_exists(self, variable):
        """Return whether a varable exists in the workspace"""

        return seamm.flowchart_variables.exists(variable)

    def delete_variable(self, variable):
        """Delete a variable in the workspace"""

        seamm.flowchart_variables.delete(variable)

    def setup_printing(self, printer):
        """Establish the handlers for printing as controlled by
        options
        """

        # Control output going to the main job printer
        # If we are in a loop, don't print to the job output, except
        # at the JOB level
        # job.setLevel(printing.NORMAL)
        # for segment in self._id:
        #     if str(segment)[0:5] == 'iter_':
        #         job.setLevel(printing.JOB)
        #         break

        # First remove an existing handlers
        self.close_printing(printer)

        # A handler for stdout
        console_handler = logging.StreamHandler()
        console_handler.setLevel(printing.JOB)
        console_handler.setFormatter(self.formatter)
        printer.addHandler(console_handler)

        # A handler for the file
        path = Path(self.directory) / "step.out"
        path.unlink(missing_ok=True)
        file_handler = logging.FileHandler(path, delay=True)
        file_handler.setLevel(printing.NORMAL)
        file_handler.setFormatter(self.formatter)
        printer.addHandler(file_handler)

        # # A handler for the job file
        # wdir = self.flowchart.root_directory
        # job_file_handler = logging.FileHandler(os.path.join(wdir, 'job.out'))
        # # job_file_handler.setLevel(printing.JOB)
        # job_file_handler.setLevel(printing.NORMAL)
        # job_file_handler.setFormatter(self.formatter)
        # printer.addHandler(job_file_handler)

    def close_printing(self, printer):
        """Close the handlers for printing, so that buffers are
        flushed, files closed, etc.
        """
        if printer is not None:
            for handler in printer.handlers:
                handler.close()
                printer.removeHandler(handler)

    def job_output(self, text):
        """Temporary!"""
        job.job(text)

    def store_results(
        self,
        configuration=None,
        data={},
        create_tables=True,
    ):
        """Store results in the database, as variables,and in tables.

        Parameters
        ----------
        configuration : molsystem._Configuration
           The configuration for storing properties in the database.
        data : dict(str, dict(str, any))
           The data resulting from running the step.
        create_tables : bool, optional
           Whether to create tables that do not yet exist, default is True.
        """
        if "results" not in self.parameters:
            return

        results = self.parameters["results"].value

        for key, value in results.items():
            if key not in data or data[key] is None:
                continue

            # The metadata describing this result
            result_metadata = self.metadata["results"][key]

            # Store the value in the database as a property.
            if "property" in value and value["property"]:
                if configuration is not None:
                    properties = configuration.properties
                    _property = value["property"].format(model=self.model)

                    self.logger.debug(
                        f"setting property '{_property}' = {data[key]} ({key=})"
                    )

                    if properties.exists(_property):
                        units = properties.units(_property)
                    else:
                        # Get the general property's info to create the model property.
                        _type, units, description = properties.metadata(
                            value["property"]
                        )
                        properties.add(
                            _property,
                            _type=_type,
                            units=units,
                            description=description.format(model=self.model),
                        )
                    # May need to convert units to those for this property.
                    if "units" in result_metadata:
                        current_units = result_metadata["units"]
                        if units != current_units:
                            if result_metadata["dimensionality"] == "scalar":
                                tmp = Q_(data[key], current_units)
                                properties.put(_property, tmp.m_as(units))
                            else:
                                factor = Q_(1, current_units).m_as(units)
                                tmp = scale(data[key], factor)
                                properties.put(
                                    _property, json.dumps(tmp, separators=(",", ":"))
                                )
                        else:
                            if result_metadata["dimensionality"] == "scalar":
                                properties.put(_property, data[key])
                            else:
                                properties.put(
                                    _property,
                                    json.dumps(data[key], separators=(",", ":")),
                                )
                    else:
                        if result_metadata["dimensionality"] == "scalar":
                            properties.put(_property, data[key])
                        else:
                            properties.put(
                                _property, json.dumps(data[key], separators=(",", ":"))
                            )

            # Store in a variable
            if "variable" in value:
                # Name of the variable
                variable = self.get_value(value["variable"])

                self.logger.debug(f"setting '{variable}' = {data[key]} (key={key})")

                # Convert the value to the requested units.
                if "units" in results[key]:
                    units = results[key]["units"]
                    if "units" in result_metadata:
                        current_units = result_metadata["units"]
                        if units != current_units:
                            tmp = Q_(data[key], current_units)
                            self.set_variable(variable, tmp.m_as(units))
                        else:
                            self.set_variable(variable, data[key])
                    else:
                        raise RuntimeError("Problem with units handling results!")
                else:
                    self.set_variable(variable, data[key])

            # Store in a table
            if "table" in value:
                tablename = value["table"]
                column = self.get_value(value["column"])
                # Does the table exist?
                if not self.variable_exists(tablename):
                    # Create the table if allowed to.
                    if create_tables:
                        table = pandas.DataFrame()
                        self.set_variable(
                            tablename,
                            {
                                "type": "pandas",
                                "table": table,
                                "defaults": {},
                                "loop index": False,
                                "current index": 0,
                                "index column": None,
                            },
                        )
                    else:
                        raise RuntimeError(
                            "Table '{}' does not exist.".format(tablename)
                        )

                table_handle = self.get_variable(tablename)
                table = table_handle["table"]

                # create the column as needed handling "key"ed columns
                if "{key}" in column:
                    if not isinstance(data[key], dict):
                        raise ValueError(
                            f"Data for a keyed column '{column}' is not a dictionary. "
                            f"{type(data[key])}"
                        )
                    for ckey, value in data[key].items():
                        keyed_column = column.replace("{key}", ckey)
                        if keyed_column not in table.columns:
                            if "units" in result_metadata:
                                units = result_metadata["units"]
                                if "units" in results[key]:
                                    units = results[key]["units"]
                                keyed_column += f" ({units})"
                        if keyed_column not in table.columns:
                            if result_metadata["dimensionality"] == "scalar":
                                kind = result_metadata["type"]
                                if kind == "boolean":
                                    default = False
                                elif kind == "integer":
                                    default = 0
                                elif kind == "float":
                                    default = np.nan
                                else:
                                    default = ""
                            else:
                                kind = "json"
                                default = ""

                            table_handle["defaults"][keyed_column] = default
                            table[keyed_column] = default

                        # Convert the value to the requested units and put in table.
                        row_index = table_handle["current index"]
                        if "units" in results[key]:
                            units = results[key]["units"]
                            if "units" in result_metadata:
                                current_units = result_metadata["units"]
                                if units != current_units:
                                    if result_metadata["dimensionality"] == "scalar":
                                        tmp = Q_(value, current_units)
                                        table.at[row_index, keyed_column] = tmp.m_as(
                                            units
                                        )
                                    else:
                                        factor = Q_(1, current_units).m_as(units)
                                        tmp = scale(value, factor)
                                        table.at[row_index, keyed_column] = json.dumps(
                                            tmp, separators=(",", ":")
                                        )
                                else:
                                    if result_metadata["dimensionality"] == "scalar":
                                        table.at[row_index, keyed_column] = value
                                    else:
                                        table.at[row_index, keyed_column] = json.dumps(
                                            value, separators=(",", ":")
                                        )
                            else:
                                raise RuntimeError(
                                    "Problem with units handling results!"
                                )
                        else:
                            if result_metadata["dimensionality"] == "scalar":
                                table.at[row_index, keyed_column] = value
                            else:
                                table.at[row_index, keyed_column] = json.dumps(
                                    value, separators=(",", ":")
                                )
                else:
                    if column not in table.columns:
                        if "units" in result_metadata:
                            units = result_metadata["units"]
                            if "units" in results[key]:
                                units = results[key]["units"]
                            column += f" ({units})"
                    if column not in table.columns:
                        if result_metadata["dimensionality"] == "scalar":
                            kind = result_metadata["type"]
                            if kind == "boolean":
                                default = False
                            elif kind == "integer":
                                default = 0
                            elif kind == "float":
                                default = np.nan
                            else:
                                default = ""
                        else:
                            kind = "json"
                            default = ""

                        table_handle["defaults"][column] = default
                        table[column] = default

                    # Convert the value to the requested units and put in table.
                    row_index = table_handle["current index"]
                    if "units" in results[key]:
                        units = results[key]["units"]
                        if "units" in result_metadata:
                            current_units = result_metadata["units"]
                            if units != current_units:
                                if result_metadata["dimensionality"] == "scalar":
                                    tmp = Q_(data[key], current_units)
                                    table.at[row_index, column] = tmp.m_as(units)
                                else:
                                    factor = Q_(1, current_units).m_as(units)
                                    tmp = scale(data[key], factor)
                                    table.at[row_index, column] = json.dumps(
                                        tmp, separators=(",", ":")
                                    )
                            else:
                                if result_metadata["dimensionality"] == "scalar":
                                    table.at[row_index, column] = data[key]
                                else:
                                    table.at[row_index, column] = json.dumps(
                                        data[key], separators=(",", ":")
                                    )
                        else:
                            raise RuntimeError("Problem with units handling results!")
                    else:
                        if result_metadata["dimensionality"] == "scalar":
                            table.at[row_index, column] = data[key]
                        else:
                            table.at[row_index, column] = json.dumps(
                                data[key], separators=(",", ":")
                            )

    def create_figure(self, title="", template="line.graph_template", module_path=None):
        """Create a new figure.

        Parameters
        ----------
        title : str, optional
            The title of the figure
        template : str, optional
            The Jinja template for the desired graph. Defaults to
            'line.graph_template'

        Returns
        -------
        seamm_util.Figure
        """

        if self._jinja_env is None:
            # The order of the loaders is important! They are searched
            # in order, so the first has precedence. This searches the
            # current package first, then looks in the main SEAMM
            # templates.
            if module_path is None:
                self.logger.info("Reading graph templates from 'seamm'")
                loaders = [jinja2.PackageLoader("seamm")]
            else:
                self.logger.info(
                    "Reading graph templates from the following modules, in order"
                )
                loaders = []
                for module in module_path:
                    paths = []
                    for p in implib.files(module):
                        if p.parent.name == "templates":
                            paths.append(p)
                            break

                    if len(paths) == 0:
                        self.logger.info(f"\t{module} -- found no templates directory")
                    else:
                        path = paths[0].locate().parent
                        self.logger.info(f"\t{ module} --> {path}")
                        loaders.append(jinja2.FileSystemLoader(path))

            self._jinja_env = jinja2.Environment(loader=jinja2.ChoiceLoader(loaders))

        figure = seamm_util.Figure(
            jinja_env=self._jinja_env, template=template, title=title
        )
        return figure

    def create_parser(self, name=None):
        """Create the parser for this node.

        All nodes have at least --log-level, which is setup here,

        Parameters
        ----------
        name : str
            The name of the parser. Defaults to a name derived from the module.

        Returns
        -------
        seamm.Node() :
            The next node in the flowchart.
        """
        if self.visited:
            return None

        self.visited = True
        result = self.next()

        if name is None:
            name = self.step_type
        parser = self.flowchart.parser

        if not parser.exists(name):
            parser.add_parser(name)

            parser.add_argument_group(
                name,
                "debugging options",
                "Options for turning on debugging output and tools",
            )

            parser.add_argument(
                name,
                "--log-level",
                group="debugging options",
                default="WARNING",
                type=str.upper,
                choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
                help=(
                    "The level of informational output, defaults to " "'%(default)s'"
                ),
            )

        return result



# -*- coding: utf-8 -*-

"""A flowchart, which is a set of nodes. There must be a single
'start' node, with other nodes connected via their ports to describe
the flowchart. There may be isolated nodes or groups of connected nodes;
however, the flow starts at the 'start' node and follows the connections,
so isolated nodes and fragments will not be executed."""

import configparser
from datetime import datetime
import hashlib
import json
import logging
from pathlib import Path
import os
import os.path
import stat

from packaging.version import Version

import seamm
import seamm_util

logger = logging.getLogger(__name__)


class Flowchart(object):
    graphics = "Tk"
    """str: The default graphics to use for display, if
    needed. Default: 'Tk'
    """

    def __init__(
        self,
        parent=None,
        data=None,
        namespace="org.molssi.seamm",
        name="",
        description="",
        directory=None,
        output="files",
        parser_name="SEAMM",
    ):
        """Initialize the flowchart.

        Parameters
        ----------
        parent : Object
            The parent of the nodes in this flowchart.
        data : dict
            An initial graph.
        namespace : str
            The namespace for locating plug-ins.
        name : str
            A name for the flowchart
        description : str
            A description of the flowcharts.
        directory : str
            The root directory for files for this flowchart.
        output : str
            Where to direct output. Currently not used.
        parser_name : str = "SEAMM"
            The name of the parser.
        """

        self.graph = seamm.Graph()
        self.parent = parent
        self.output = output  # Where to print output, files, stdout, both
        self.metadata = {}
        self.reset_metadata(title=name, description=description)
        self._parser_name = parser_name

        # Setup the plugin handling
        self.plugin_manager = seamm.PluginManager(namespace)

        # and make sure that the start node exists
        self.add_node(seamm.StartNode(flowchart=self))

        # And the root directory
        self.root_directory = directory

        # And the parser associated with this flowchart
        self._parser = None

    def __iter__(self):
        return self.graph.__iter__()

    @property
    def is_development(self):
        """Check if any of nodes are development versions."""
        for node in self:
            if Version(node.version).is_prerelease:
                return True
        return False

    @property
    def root_directory(self):
        """The root directory for files, etc for this flowchart"""
        if self._root_directory is None:
            self._root_directory = os.path.join(
                os.getcwd(), datetime.now().isoformat(sep="_", timespec="seconds")
            )
        return self._root_directory

    @root_directory.setter
    def root_directory(self, value):
        self._root_directory = value

    @property
    def output(self):
        """Where to print output:
        files:  to files in subdirectories
        stdout: to standard output (for intereactive use)
        both:   to both files and standard output
        """
        return self._output

    @output.setter
    def output(self, value):
        if value in ("files", "stdout", "both"):
            self._output = value
        else:
            raise RuntimeError(
                "flowchart.output must be one of 'files', 'stdout', or 'both'"
                ", not '{}'".format(value)
            )

    @property
    def parser(self):
        """The SEAMM parser associated with this flowchart."""
        if self._parser is None:
            self._parser = seamm_util.seamm_parser(self._parser_name)
        return self._parser

    @parser.setter
    def parser(self, value):
        self._parser = value

    # -------------------------------------------------------------------------
    # Node creation and deletion
    # -------------------------------------------------------------------------

    def create_node(self, extension_name):
        """Create a new node given the extension name"""
        plugin = self.plugin_manager.get(extension_name)
        node = plugin.create_node(flowchart=self, extension=extension_name)
        node.parent = self.parent
        return node

    def add_node(self, n, **attr):
        """Add a single node n, ensuring that it knows the flowchart"""
        n.flowchart = self
        return self.graph.add_node(n, **attr)

    def remove_node(self, node):
        """Delete a node from the flowchart, and from the graphics if
        necessary
        """

        # Remove the drawing of the node
        node.undraw()

        # and any edges, including graphics if appropriate
        node.remove_edge("all")

        # and the node
        self.graph.remove_node(node)

    # -------------------------------------------------------------------------
    # Finding nodes
    # -------------------------------------------------------------------------

    def tag_exists(self, tag):
        """Check if the node with a given tag exists. A tag is a string like
        'node=<uuid>', where <uuid> is the unique integer id for the node.
        """

        for node in self:
            if node.tag == tag:
                return True
        return False

    def get_node(self, uuid):
        """Return the node with a given uuid"""
        if isinstance(uuid, int):
            uuid = str(uuid)
        for node in self:
            if str(node.uuid) == uuid:
                return node
        return None

    def get_nodes(self):
        "Return a list of all the nodes in the traversal."
        logger.debug("Getting nodes")

        # Reset the visited flag to check for loops
        self.reset_visited()

        # Get the start node
        next_node = self.get_node("1")

        # And traverse the nodes.
        nodes = []
        while next_node:
            nodes.append(next_node)
            next_node = next_node.next()
        logger.debug("Finished getting nodes")
        return nodes

    def last_node(self, node="1"):
        """Find the last node walking down the main execution path
        from the given node, which defaults to the start node"""

        if isinstance(node, str):
            node = self.get_node(node)

        for edge in self.graph.edges(node, direction="out"):
            if edge.edge_type == "execution":
                return self.last_node(edge.node2)

        return node

    def clear(self, all=False):
        """Override the underlying clear() to ensure that the start node is present"""
        self.graph.clear()

        # and make sure that the start node exists
        if not all:
            start_node = seamm.StartNode(flowchart=self)
            self.add_node(start_node)

    def list_nodes(self):
        """List the nodes, for debugging"""
        result = []
        for node in self:
            result.append(node.__class__.__name__ + " {}".format(node))
        return result

    # -------------------------------------------------------------------------
    # Traversal
    # -------------------------------------------------------------------------

    def reset_visited(self):
        """Reset the 'visited' flag, which is used to detect
        loops during traversals
        """

        for tmp in self:
            tmp.visited = False

    def set_ids(self, node_id=()):
        """Sequentially number all nodes, and subnodes"""
        logger.debug("Setting ids")

        # Clear all ids
        for node in self:
            node.reset_id()

        # Reset the visited flag to check for loops
        self.reset_visited()

        # Get the start node
        next_node = self.get_node("1")

        # And traverse the nodes.
        n = 0
        while next_node:
            next_node = next_node.set_id((*node_id, str(n)))
            n += 1
        logger.debug("Finished setting ids")

    def create_parsers(self):
        """Create the argument parsers for the nodes."""
        logger.debug("Creating argument parsers.")
        # Reset the visited flag to check for loops
        self.reset_visited()

        # Get the start node
        next_node = self.get_node("1")

        # And traverse the nodes.
        while next_node:
            next_node = next_node.create_parser()
        logger.debug("Finished creating argument parsers.")

    def set_log_level(self, options):
        """Set the log level for each node based on the options"""
        logger.debug("Setting the log-level")

        for node in self:
            step_type = node.step_type
            logger.debug(f"    checking for node type {step_type}")
            if step_type in options and "log_level" in options[step_type]:
                logger.debug(f"      log_level = {options[step_type]['log_level']}")
                try:
                    node.logger.setLevel(options[step_type]["log_level"])
                    logger.debug("        set!")
                except Exception as e:
                    print(f"Exception {type(e)}: {e}")
        logger.debug("Finished setting node log levels.")

    # -------------------------------------------------------------------------
    # Strings, reading and writing
    # -------------------------------------------------------------------------

    def digest(self, strict=False):
        """Generate a unique hash key for this flowchart.

        Parameters
        ----------
        strict: bool
            Whether to include version information. Default: False

        Returns
        -------
        string
        """
        hasher = hashlib.sha256()

        # Hash the nodes and edges in the graph by traversing the graph

        # Reset the visited flag to check for loops
        self.reset_visited()

        # Get the start node
        next_node = self.get_node("1")

        # And traverse the nodes.
        while next_node:
            if next_node.visited:
                break
            next_node.visited = True
            hasher.update(bytes(next_node.digest(strict=strict), "utf-8"))
            next_node = next_node.next()

        return hasher.hexdigest()

    def to_json(self):
        """Ufff. Turn ourselves into JSON"""
        return json.dumps(self.to_dict())

    def to_dict(self):
        """Serialize the graph and everything it contains in a dict"""

        data = {
            "item": "object",
            "module": self.__module__,
            "class": self.__class__.__name__,
            "extension": None,
        }

        nodes = data["nodes"] = []
        for node in self:
            nodes.append(node.to_dict())

        edges = data["edges"] = []
        for edge in self.graph.edges():
            attr = {}
            for key in edge:
                if key not in ("node1", "node2", "edge_type", "edge_subtype"):
                    attr[key] = edge[key]
            edges.append(
                {
                    "item": "edge",
                    "node1": edge.node1.uuid,
                    "node2": edge.node2.uuid,
                    "edge_type": edge.edge_type,
                    "edge_subtype": edge.edge_subtype,
                    "attributes": attr,
                }
            )

        return data

    def from_dict(self, data):
        """recreate the flowchart from the dict. Inverse of to_dict."""
        if "class" not in data:
            raise RuntimeError("There is no class information in the data!")
        if data["class"] != self.__class__.__name__:
            raise RuntimeError(
                "The dictionary does not contain a flowchart!"
                " It contains a {} class".format(data["class"])
            )

        self.clear()

        # Recreate the nodes
        for node in data["nodes"]:
            if node["class"] == "StartNode":
                new_node = self.get_node("1")
                new_node.from_dict(node)
                continue

            logger.debug("creating {} node".format(node["extension"]))
            plugin = self.plugin_manager.get(node["extension"])
            logger.debug("  plugin object: {}".format(plugin))

            # Recreate the node
            new_node = plugin.create_node(flowchart=self, extension=node["extension"])
            new_node.parent = self.parent

            # set uuid to correct value
            new_node._uuid = node["attributes"]["_uuid"]

            # and add to the flowchart
            self.add_node(new_node)

            new_node.from_dict(node)

            logger.debug("adding nodes: nodes:\n\t" + "\n\t".join(self.list_nodes()))

        # and the edges connecting them
        for edge in data["edges"]:
            node1 = self.get_node(edge["node1"])
            node2 = self.get_node(edge["node2"])
            self.add_edge(
                node1,
                node2,
                edge_type=edge["edge_type"],
                edge_subtype=edge["edge_subtype"],
                **edge["attributes"],
            )

            logger.debug("Adding edges, nodes:\n\t" + "\n\t".join(self.list_nodes()))

    def write(self, filename):
        """Write the serialized form to disk"""
        with open(filename, "w") as fd:
            fd.write(self.to_text())

        logger.info(f"Wrote flowchart to {filename}")

        permissions = stat.S_IMODE(os.lstat(filename).st_mode)
        os.chmod(filename, permissions | stat.S_IXUSR | stat.S_IXGRP)

    def to_text(self):
        """Return the text for the flowchart.

        This is the representation written to disk, submitted
        as jobs, etc. There are two header lines followed by json
        representing the flowchart.

        Returns
        -------
        str : the text representation.
        """
        text = "#!/usr/bin/env run_flowchart\n"
        text += "!MolSSI flowchart 2.0\n"
        text += "#metadata\n"
        self.metadata["sha256"] = self.digest()
        self.metadata["sha256_strict"] = self.digest(strict=True)
        text += json.dumps(self.metadata, indent=4)
        text += "\n"
        text += "#flowchart\n"
        text += json.dumps(self.to_dict(), indent=4, cls=seamm_util.JSONEncoder)
        text += "\n"
        text += "#end\n"

        return text

    def from_text(self, text):
        """Recreate the flowchart from text"""
        lines = iter(text.splitlines())

        line = next(lines)
        # There may be exec magic as first line
        if line[0:2] == "#!":
            line = next(lines)
        if line[0:7] != "!MolSSI":
            raise RuntimeError("File is not a MolSSI file! -- " + line)
        tmp = line.split()
        if len(tmp) < 3:
            raise RuntimeError("File is not a proper MolSSI file! -- " + line)
        if tmp[1] != "flowchart":
            raise RuntimeError("File is not a flowchart! -- " + line)
        flowchart_version = tmp[2]
        version = Version(flowchart_version)
        logger.info(f"Reading flowchart version {flowchart_version}")

        if version < Version("2.0"):
            self.metadata = {}
            rest = "\n".join([x for x in lines])
            data = json.loads(rest, cls=seamm_util.JSONDecoder)
        else:
            sections = {}
            in_section = False
            for line in lines:
                if line.strip() == "":
                    continue
                if line[0] == "#":
                    section = line.strip()[1:]
                    if section == "end":
                        in_section = False
                    else:
                        tmp = sections[section] = []
                        in_section = True
                    continue
                elif in_section:
                    tmp.append(line)

            self.metadata = json.loads("\n".join(sections["metadata"]))
            data = json.loads("\n".join(tmp), cls=seamm_util.JSONDecoder)

        if "class" not in data or data["class"] != "Flowchart":
            raise RuntimeError("Text does not contain a flowchart")

        self.from_dict(data)

    def read(self, filename):
        """Recreate the flowchart from the serialized form on disk"""
        with open(filename, "r") as fd:
            text = fd.read()

        self.from_text(text)

    def reset_metadata(self, **kwargs):
        """Setup the metadata initially."""
        self.metadata = {
            "title": "",
            "description": "",
            "keywords": [],
            "creators": [],
            "grants": [],
        }
        # See if the user info is in the .seammrc file...
        path = Path("~/.seammrc").expanduser()
        if path.exists:
            config = configparser.ConfigParser()
            config.read(path)
            if "USER" in config:
                user = config["USER"]
                if "name" in user:
                    author = {"name": user["name"]}
                    for key in ("orcid", "affiliation"):
                        if key in user:
                            author[key] = user[key]
                    self.metadata["creators"].append(author)
                if "grants" in user:
                    self.metadata["grants"] = user["grants"].split()

        self.metadata.update(kwargs)

    # -------------------------------------------------------------------------
    # Edges between nodes
    # -------------------------------------------------------------------------

    def edges(self, node=None, direction="both"):
        return self.graph.edges(node, direction)

    def add_edge(self, u, v, edge_type=None, edge_subtype="next", **attr):
        return self.graph.add_edge(u, v, edge_type, edge_subtype, **attr)

    def print_edges(self, event=None):
        """Print all the edges. Useful for debugging!"""

        print("All edges in flowchart")
        for edge in self.edges():
            # print('   {}'.format(edge))
            print(
                "   {} {} {} {} {} {}".format(
                    edge.node1.tag,
                    edge.anchor1,
                    edge.node2.tag,
                    edge.anchor2,
                    edge.edge_type,
                    edge.edge_subtype,
                )
            )

    # -------------------------------------------------------------------------
    # Printing
    # -------------------------------------------------------------------------



import argparse
import locale
import logging
import os
import platform
import Pmw
import subprocess
import sys
import tkinter as tk
import tkinter.font as tkFont

import seamm
import seamm_util

logger = logging.getLogger(__name__)
dbg_level = 30

standard_fonts = {"scale": 1.0}


def decrease_font_size(event=None, factor=1.3):
    """Uniformly decrease the font sizes."""
    global standard_fonts
    scale = standard_fonts["scale"] / factor
    standard_fonts["scale"] = scale
    for font_name, data in standard_fonts.items():
        if font_name == "scale":
            continue
        size = data["initial size"]
        new_size = int(size * scale)
        if new_size == size:
            new_size -= 1
        if new_size < 8:
            new_size = 8
        font = tkFont.nametofont(font_name)
        font.config(size=new_size)
        data["current size"] = new_size


def increase_font_size(event=None, factor=1.3):
    """Uniformly increase the font sizes."""
    global standard_fonts
    scale = standard_fonts["scale"] * factor
    standard_fonts["scale"] = scale
    for font_name, data in standard_fonts.items():
        if font_name == "scale":
            continue
        size = data["initial size"]
        new_size = int(size * scale + 0.5)
        if new_size == size:
            new_size += 1
        font = tkFont.nametofont(font_name)
        font.config(size=new_size)
        data["current size"] = new_size


def raise_app(root: tk):
    root.attributes("-topmost", True)
    if platform.system() == "Darwin":
        tmpl = (
            'tell application "System Events" to set frontmost '
            "of every process whose unix id is {} to true"
        )
        script = tmpl.format(os.getpid())
        subprocess.check_call(["/usr/bin/osascript", "-e", script])
    root.after(100, lambda: root.attributes("-topmost", False))


def reset_font_size():
    """Reset the font sizes."""
    global standard_fonts
    scale = 1.0
    standard_fonts["scale"] = scale
    for font_name, data in standard_fonts.items():
        if font_name == "scale":
            continue
        size = data["initial size"]
        font = tkFont.nametofont(font_name)
        font.config(size=size)
        data["current size"] = size


def flowchart():
    """The standalone flowchart app"""
    global app_name
    app_name = "MolSSI SEAMM"
    global dbg_level
    global standard_fonts

    parser = seamm_util.getParser("SEAMM GUI")
    parser.add_parser(
        "SEAMM",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        prog=sys.argv[0],
    )
    parser.add_argument(
        "SEAMM",
        "--version",
        action="version",
        version=f"SEAMM version {seamm.__version__}",
    )
    parser.add_argument(
        "SEAMM",
        "flowcharts",
        nargs="*",
        default=[],
        help="flowcharts to open initially",
    )

    # Debugging options
    parser.add_argument_group(
        "SEAMM",
        "debugging options",
        "Options for turning on debugging output and tools",
    )
    parser.add_argument(
        "SEAMM",
        "--log-level",
        group="debugging options",
        default="WARNING",
        type=str.upper,
        choices=["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        help="The level of informational output, default: '%(default)s'",
    )

    parser.add_parser("SEAMM GUI")

    parser.add_argument(
        "SEAMM GUI",
        "--font-scale",
        default=1.0,
        help="scale factor for the fonts",
    )

    parser.parse_args()
    options = parser.get_options("SEAMM")
    gui_options = parser.get_options("SEAMM GUI")

    if "log_level" in options:
        logging.basicConfig(level=options["log_level"])

    ##################################################
    # Initialize Tk
    ##################################################
    root = tk.Tk()
    Pmw.initialise(root)

    # Capture the initial font information
    for font_name in tkFont.names():
        font = tkFont.nametofont(font_name)
        standard_fonts[font_name] = {
            "initial size": font.cget("size"),
            "current size": font.cget("size"),
        }

    try:
        font_scale = float(gui_options["font_scale"])
        if font_scale != 1.0:
            increase_font_size(factor=font_scale)
    except Exception:
        pass

    ##############################################################
    # Create the various objects that we need: the model, the view
    # and the data
    ##############################################################

    flowchart = seamm.Flowchart()
    tk_flowchart = seamm.TkFlowchart(master=root, flowchart=flowchart)
    # The data is implicitly initialized to none...

    ##################################################
    # Initialize the rest of the GUI, such as menus
    ##################################################
    logger.debug("Initializing the rest of the GUI")

    root.title(app_name)

    # The menus
    menu = tk.Menu(root)

    # Set the about and preferences menu items on Mac
    if sys.platform.startswith("darwin"):
        app_menu = tk.Menu(menu, name="apple")
        menu.add_cascade(menu=app_menu)

        app_menu.add_command(label="About " + app_name, command=tk_flowchart.about)
        app_menu.add_separator()
        root.createcommand("tk::mac::ShowPreferences", tk_flowchart.preferences)
        root.createcommand("tk::mac::OpenDocument", tk_flowchart.open_file)
        CmdKey = "Command-"
    else:
        CmdKey = "Control-"

    root.config(menu=menu)
    filemenu = tk.Menu(menu)
    menu.add_cascade(label="File", menu=filemenu)
    filemenu.add_command(
        label="New", command=tk_flowchart.new_file, accelerator=CmdKey + "N"
    )
    filemenu.add_command(
        label="Open...", command=tk_flowchart.flowchart_search, accelerator=CmdKey + "O"
    )
    filemenu.add_command(
        label="Save...", command=tk_flowchart.save, accelerator=CmdKey + "S"
    )
    filemenu.add_command(label="Save as...", command=tk_flowchart.save_file)
    filemenu.add_command(label="Publish...", command=tk_flowchart.publish)
    filemenu.add_separator()
    filemenu.add_command(
        label="Run", command=tk_flowchart.run, accelerator=CmdKey + "R"
    )

    # Control debugging info
    filemenu.add_separator()
    debug_menu = tk.Menu(menu)
    filemenu.add_cascade(label="Debug", menu=debug_menu)
    debug_menu.add_radiobutton(
        label="normal",
        value=30,
        variable=dbg_level,
        command=lambda arg0=30: handle_dbg_level(arg0),
    )
    debug_menu.add_radiobutton(
        label="info",
        value=20,
        variable=dbg_level,
        command=lambda arg0=20: handle_dbg_level(arg0),
    )
    debug_menu.add_radiobutton(
        label="debug",
        value=10,
        variable=dbg_level,
        command=lambda arg0=10: handle_dbg_level(arg0),
    )

    # Exiting
    filemenu.add_separator()
    filemenu.add_command(label="Exit", command=root.quit)

    # Edit menu
    editmenu = tk.Menu(menu)
    menu.add_cascade(label="Edit", menu=editmenu)
    editmenu.add_command(label="Description...", command=tk_flowchart.properties)
    editmenu.add_command(
        label="Clean layout",
        command=tk_flowchart.clean_layout,
        accelerator=CmdKey + "l",
    )

    # View menu
    viewmenu = tk.Menu(menu)
    menu.add_cascade(label="View", menu=viewmenu)
    viewmenu.add_command(
        label="Increase font size", command=increase_font_size, accelerator=CmdKey + "+"
    )
    viewmenu.add_command(
        label="Decrease font size", command=decrease_font_size, accelerator=CmdKey + "-"
    )
    viewmenu.add_command(label="Reset font size", command=reset_font_size)

    # Help menu
    helpmenu = tk.Menu(menu)
    menu.add_cascade(label="Help", menu=helpmenu)
    if sys.platform.startswith("darwin"):
        root.createcommand("tk::mac::ShowHelp", tk_flowchart.help)

    root.bind_all("<" + CmdKey + "N>", tk_flowchart.new_file)
    root.bind_all("<" + CmdKey + "n>", tk_flowchart.new_file)
    root.bind_all("<" + CmdKey + "O>", tk_flowchart.flowchart_search)
    root.bind_all("<" + CmdKey + "o>", tk_flowchart.flowchart_search)
    root.bind_all("<" + CmdKey + "R>", tk_flowchart.run)
    root.bind_all("<" + CmdKey + "r>", tk_flowchart.run)
    root.bind_all("<" + CmdKey + "S>", tk_flowchart.save)
    root.bind_all("<" + CmdKey + "s>", tk_flowchart.save)
    root.bind_all("<" + CmdKey + "L>", tk_flowchart.clean_layout)
    root.bind_all("<" + CmdKey + "l>", tk_flowchart.clean_layout)
    root.bind_all("<" + CmdKey + "plus>", increase_font_size)
    root.bind_all("<" + CmdKey + "equal>", increase_font_size)
    root.bind_all("<" + CmdKey + "minus>", decrease_font_size)

    # Work out and set the window size to nicely fit the screen
    sw = root.winfo_screenwidth()
    sh = root.winfo_screenheight()
    w = int(0.9 * sw)
    h = int(0.8 * sh)
    x = int(0.1 * sw / 2)
    y = int(0.2 * sh / 2)

    root.geometry("{}x{}+{}+{}".format(w, h, x, y))

    logger.debug("Finished initializing the rest of the GUI, drawing window")

    # Draw the flowchart
    tk_flowchart.draw()

    logger.debug("SEAMM has been drawn. Now raise it to the top")

    # bring it to the top of all windows
    root.lift()
    raise_app(root)

    # Check to see if the command line has flowcharts to open
    if len(options["flowcharts"]) > 0:
        logger.debug("open the following flowcharts:")
        if len(options["flowcharts"]) > 1:
            raise RuntimeError("Currently handle only one flowchart at a time")
        for filename in options["flowcharts"]:
            tk_flowchart.open(filename)

    logger.debug("and now enter the event loop")

    # enter the event loop
    root.mainloop()


def handle_dbg_level(level):
    global dbg_level

    dbg_level = level
    logging.getLogger().setLevel(dbg_level)


if __name__ == "__main__":
    locale.setlocale(locale.LC_ALL, "")

    flowchart()



# -*- coding: utf-8 -*-

"""A singleton to ensure the ~.seammrc file is always up-to-date."""

import configparser
from pathlib import Path

# Used in parser getters to indicate the default behaviour when a specific
# option is not found it to raise an exception. Created to enable `None' as
# a valid fallback value.
_UNSET = object()


class Singleton(object):
    _instances = {}

    def __new__(class_, *args, **kwargs):
        if class_ not in class_._instances:
            class_._instances[class_] = super(Singleton, class_).__new__(
                class_, *args, **kwargs
            )
        return class_._instances[class_]


class SEAMMrc(Singleton):
    def __init__(self, path="~/.seammrc"):
        self._config = configparser.ConfigParser()
        self.path = Path(path).expanduser()

        # Create the file if it doesn't exist
        if self.path.exists():
            self._config.read(self.path)
        else:
            self.path.parent.mkdir(parents=True, exist_ok=True)
            self._save()

        # Check the version and upgrade if necessary
        if "VERSION" not in self._config:
            # Rename all sections as Dashboards
            for section in self._config.sections():
                tmp = {}
                for key, value in self._config[section].items():
                    tmp[key] = value
                self._config.remove_section(section)
                self._config[f"Dashboard: {section}"] = tmp
            self._config["VERSION"] = {"file": "1.0"}
            self._save()

    def __getitem__(self, key):
        raise NotImplementedError("Please use get/set")

    def __setitem__(self, key, value):
        raise NotImplementedError("Please use get/set")

    def __delitem__(self, key):
        del self._config[key]
        self._save()

    def __contains__(self, key):
        return key in self._config

    def __len__(self):
        return len(self._config)

    def __iter__(self):
        return self._config.__iter__()

    def defaults(self):
        return self._config.defaults()

    def sections(self):
        return self._config.sections()

    def add_section(self, section):
        self._config.add_section(section)
        self._save()

    def has_section(self, section):
        return self._config.has_section(section)

    def options(self, section):
        return self._config.options(section)

    def has_option(self, section, option):
        return self._config.has_option(section, option)

    def get(self, section, option, raw=False, vars=None, fallback=_UNSET):
        return self._config.get(section, option, raw=raw, vars=vars, fallback=fallback)

    def getint(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
        return self._config.getint(
            section, option, raw=raw, vars=vars, fallback=fallback
        )

    def getfloat(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
        return self._config.getfloat(
            section, option, raw=raw, vars=vars, fallback=fallback
        )

    def getboolean(self, section, option, *, raw=False, vars=None, fallback=_UNSET):
        return self._config.getboolean(
            section, option, raw=raw, vars=vars, fallback=fallback
        )

    def items(self, section=_UNSET, raw=False, vars=None):
        return self._config.items(section=section, raw=raw, vars=vars)

    def set(self, section, option, value):
        self._config.set(section, option, value)
        self._save()

    def remove_option(self, section, option):
        self._config.remove_option(section, option)
        self._save()

    def remove_section(self, section):
        self._config.remove_section(section)
        self._save()

    def _save(self):
        with open(self.path, "w") as fd:
            # Added commented sections if they don't exist
            if "USER" not in self:
                fd.write(
                    """
# [USER]
# Default user and grant information for flowcharts

# name = Last, First
# ORCID = xxxx-xxxx-xxxx-xxxx
# affiliation = Your instititution
# grants = <as DOIs like Zenodo uses, e.g 10.13039/100000001::2136142 10.13...>
"""
                )
            if "ZENODO" not in self:
                fd.write(
                    """
# [ZENODO]
# API token for Zenodo

# token = xxxx....
"""
                )
            if "SANDBOX" not in self:
                fd.write(
                    """
# [SANDBOX]
# API token for Zenodo's sandbox

# token = xxxxx....
"""
                )

            # And write the config file data
            self._config.write(fd)

    def re_read(self):
        self._config.read(self.path)



# -*- coding: utf-8 -*-

"""A node to join together parallel flows in a flowchart"""

import seamm
import seamm_util.printing as printing
from seamm_util.printing import FormattedText as __
import logging

logger = logging.getLogger(__name__)
job = printing.getPrinter()
printer = printing.getPrinter("join")


class Join(seamm.Node):
    def __init__(self, flowchart=None, extension="Join"):
        """Initialize a node for joining the flow together again

        Keyword arguments:
        """
        logger.debug("Constructing join node {}".format(self))
        super().__init__(flowchart=flowchart, title="Join", extension=extension)

    @property
    def version(self):
        """The semantic version of this module."""
        return seamm.__version__

    @property
    def git_revision(self):
        """The git version of this module."""
        return seamm.__git_revision__

    def description_text(self, P=None):
        """Return a short description of this step.

        Return a nicely formatted string describing what this step will
            do.

        Keyword arguments:
            P: a dictionary of parameter values, which may be variables
                or final values. If None, then the parameters values will
                be used as is.
        """

        # if not P:
        #     P = self.parameters.values_to_dict()

        text = ""
        text += "Join threads together"

        return self.header + "\n" + __(text, indent=4 * " ").__str__()



# -*- coding: utf-8 -*-

"""The GUI for opening flowcharts."""

import collections.abc
import datetime
import json
import logging
from pathlib import Path
import pprint  # noqa: F401
import tkinter as tk
from tkinter.scrolledtext import ScrolledText
import tkinter.ttk as ttk

import dateutil
import Pmw

from .dashboard_handler import DashboardHandler
from .seammrc import SEAMMrc
from seamm_dashboard_client import DashboardConnectionError
import seamm_util
import seamm_widgets as sw

logger = logging.getLogger(__name__)
# logger.setLevel("DEBUG")

# "operators": (
#     "must be",
#     "must be like",
#     "must contain",
#     "must contain like",
#     "must not be",
#     "must not be like",
#     "must not contain",
#     "must not contain like",
#     "may be",
#     "may be like",
#     "may contain",
#     "must contain like",
# ),

zenodo_fields = {
    "any field": {
        "field": "",
        "operators": (
            "contains",
            "contains like",
        ),
    },
    "author name": {
        "field": "creators.name:",
        "operators": (
            "contains",
            "contains like",
        ),
    },
    "author orcid": {"field": "creators.orcid:", "operators": ("is", "is not")},
    "author affiliation": {
        "field": "creators.affiliation:",
        "operators": (
            "contains",
            "contains like",
        ),
    },
    "community": {
        "field": "communities:",
        "operators": ("is", "is like", "is not", "is not like"),
    },
    "date": {"field": "created", "operators": ("after", "before", "on", "between")},
    "description": {
        "field": "description:",
        "operators": (
            "contains",
            "contains like",
        ),
    },
    "keyword": {
        "field": "keywords:",
        "operators": (
            "contains",
            "contains like",
        ),
    },
    "title": {
        "field": "title:",
        "operators": (
            "contains",
            "contains like",
        ),
    },
}


class TkOpen(collections.abc.MutableMapping):
    def __init__(self, toplevel):
        self._widgets = {"toplevel": toplevel}

        self._data = dict()
        self.zenodo = None
        self._dashboard_handler = None
        self.dashboards = None
        self.current_dashboard = None
        self.current_project = "any"
        self._projects = None
        self.config = SEAMMrc()
        if "SEAMM open" not in self.config:
            self.config.add_section("SEAMM open")

    # Provide dict like access to the widgets to make
    # the code cleaner

    def __getitem__(self, key):
        """Allow [] access to the widgets!"""
        return self._widgets[key]

    def __setitem__(self, key, value):
        """Allow x[key] access to the data"""
        self._widgets[key] = value

    def __delitem__(self, key):
        """Allow deletion of keys"""
        if key in self._widgets:
            self._widgets[key].destroy()
        del self._widgets[key]

    def __iter__(self):
        """Allow iteration over the object"""
        return iter(self._widgets)

    def __len__(self):
        """The len() command"""
        return len(self._widgets)

    @property
    def dashboard_handler(self):
        "The connection to the dashboards."
        if self._dashboard_handler is None:
            self._dashboard_handler = DashboardHandler()
        return self._dashboard_handler

    def clear_tree(self):
        """Remove any contents from the tree."""
        tree = self["tree"]
        children = tree.get_children()
        if len(children) > 0:
            tree.delete(*children)
        self._data = {}

    def create_dialog(self):
        """Create the dialog for opening."""
        if "dialog" in self:
            return

        self["dialog"] = d = Pmw.Dialog(
            self["toplevel"],
            buttons=("Open", "Cancel"),
            master=self["toplevel"],
            title="Open Flowchart",
        )
        d.withdraw()

        frame = self["frame"] = d.interior()

        w = self["what"] = sw.LabeledCombobox(
            frame,
            labeltext="Open a",
            values=("flowchart",),
            state="readonly",
        )
        w.set("flowchart")

        if not self.config.has_option("SEAMM open", "source"):
            self.config.set("SEAMM open", "source", "local files")
            path = Path("~/SEAMM/flowcharts").expanduser()
            path.mkdir(parents=True, exist_ok=True)
        source = self.config.get("SEAMM open", "source")

        w = self["source"] = sw.LabeledCombobox(
            frame,
            labeltext="from",
            values=("local files", "previous jobs", "Zenodo", "Zenodo sandbox"),
            state="readonly",
        )
        w.set(source)

        # For local files, a directory to start from
        if not self.config.has_option("SEAMM open", "directory"):
            self.config.set(
                "SEAMM open", "directory", json.dumps(["~/SEAMM/flowcharts"])
            )
            path = Path("~/SEAMM/flowcharts").expanduser()
            path.mkdir(parents=True, exist_ok=True)
        directories = json.loads(self.config.get("SEAMM open", "directory"))
        directory = directories[0]

        w = self["directory"] = sw.LabeledCombobox(
            frame,
            labeltext="directory",
            values=directories,
        )
        w.set(directory)

        self["get directory"] = ttk.Button(
            frame, text="...", command=self.directory_cb, width=3
        )

        # From dashboard
        if not self.config.has_option("SEAMM open", "dashboard"):
            self.config.set("SEAMM open", "dashboard", "")
        w = self["dashboard"] = sw.LabeledCombobox(
            frame,
            labeltext="Dashboard",
        )
        if not self.config.has_option("SEAMM open", "project"):
            self.config.set("SEAMM open", "project", "any")
        w = self["project"] = sw.LabeledCombobox(
            frame,
            labeltext="Project",
        )

        self["criteria"] = sw.SearchCriteria(
            frame,
            text="Show flowcharts where",
            labelanchor=tk.NW,
            inclusiontext="",
            inclusionvalues=(
                "must have",
                "must not have",
                "may have",
                "and",
                "or",
                "(",
                ")",
                "ignore",
            ),
            operatorvalues=(
                "contains",
                "contains like",
            ),
            fieldvalues=[*zenodo_fields.keys()],
            two_values=("between",),
            command=self.zenodo_callback,
        )

        self["search"] = ttk.Button(frame, text="Search", command=self.search_cb)
        w = self["tree"] = ttk.Treeview(frame, selectmode="browse")
        w.bind("<ButtonRelease-1>", self.select_record)
        w.bind(
            "<Double-Button-1>", lambda e=None: self["dialog"].deactivate(result="Open")
        )
        # Add scrollbars
        self["tree ysb"] = ttk.Scrollbar(frame, orient="vertical", command=w.yview)
        self["tree xsb"] = ttk.Scrollbar(frame, orient="horizontal", command=w.xview)
        w.configure(yscroll=self["tree ysb"].set, xscroll=self["tree xsb"].set)

        detail = self["detail"] = ttk.LabelFrame(
            frame,
            borderwidth=5,
            relief=tk.SUNKEN,
            text="Flowchart Details",
            labelanchor=tk.N,
        )

        row = 0
        w = self["title"] = ttk.Label(detail)
        w.grid(row=row, column=0)
        row += 1
        w = self["description"] = ScrolledText(
            detail, wrap=tk.WORD, font=("Helvetica",), height=6
        )
        w.grid(row=row, column=0, sticky=tk.NSEW)
        detail.rowconfigure(row, weight=1)
        detail.columnconfigure(0, weight=1)
        row += 1
        w = self["version"] = ttk.Label(detail)
        w.grid(row=row, column=0, sticky=tk.W)

        for item in ("what", "source"):
            self[item].bind("<<ComboboxSelected>>", self.reset_dialog)

        for item in ("directory",):
            self[item].bind("<<ComboboxSelected>>", self.reset_tree)
            self[item].bind("<Return>", self.reset_tree)
            self[item].bind("<FocusOut>", self.reset_tree)

        for item in ("dashboard", "project"):
            self[item].bind("<<ComboboxSelected>>", self.update_dashboard)

        # Fill the window with the dialog!
        swidth = frame.winfo_screenwidth()
        sheight = frame.winfo_screenheight()
        d.geometry(f"{int(0.8 * swidth)}x{int(0.8 * sheight)}")

    def directory_cb(self, event=None):
        """Invoked by the ... button to get new directory."""
        d = self["directory"]
        current = Path(d.get()).expanduser().resolve()
        if not current.exists() or not current.is_dir():
            current = Path("~").expanduser()
        directory = tk.filedialog.askdirectory(
            parent=self["toplevel"],
            title="Change directory",
            initialdir=current,
            mustexist=True,
        )

        if directory is None or current.samefile(directory):
            return

        d.set(directory)
        self.reset_tree()

    def fill_tree(self, job_list):
        """Fill the tree with a job list

        The job list looks is a list of Job objects that are dicts like::

            {
                'description': 'test of api',
                'finished': '2022-02-24 10:06',
                'flowchart_id': '1',
                'group': None,
                'group_id': None,
                'id': 18,
                'last_update': '2022-02-24 10:05',
                'owner': 'psaxe',
                'owner_id': 2,
                'parameters': {
                     'cmdline': ['job:data/Users_psaxe_SEAMM_data_TiO2--anatase.cif']
                },
                'path': '/Users/psaxe/SEAMM_DEV/Jobs/projects/default/Job_000018',
                'projects': [{'id': 1, 'name': 'default'}],
                'started': '2022-02-24 10:06',
                'status': 'finished',
                'submitted': '2022-02-24 10:05',
                'title': 'test of api'
            }

        Parameters
        ----------
        job_list : [seamm_dashboard_client._Job]
            List of Job objects contain info about the jobs
        """
        # Remove any current data
        self.clear_tree()
        self._data = {}
        tree = self["tree"]
        for job in job_list:
            iid = tree.insert("", "end", text=f"Job_{job['id']:0>6d}: {job['title']}")
            self._data[iid] = job

    def insert_node(self, parent, text, path, open=False):
        """Insert a new node in the tree, corresponding to a file or directory.

        Parameters
        ----------
        parent : str
            The parent node in the tree, "" for toplevel.
        text : str
            The text to display for the node.
        path : pathlib.Path
            The absolute path of the file or directory.
        """
        tree = self["tree"]
        node = tree.insert(parent, "end", text=text, open=open)
        self._data[node] = path
        if path.is_dir():
            tree.insert(node, "end")
        return node

    def open_node(self, event=None):
        tree = self["tree"]
        node = tree.focus()
        path = self._data.get(node, None)
        if path is not None:
            tree.delete(*tree.get_children(node))
            for p in sorted(path.iterdir(), key=lambda p: p.name):
                if p.is_dir() or p.suffix == ".flow":
                    self.insert_node(node, p.name, p)

    def reset_dialog(self, event=None):
        """Layout the widgets in the dialog according to the parameters."""

        frame = self["frame"]
        what = self["what"].get()
        source = self["source"].get()

        # Temporary
        if what != "flowchart":
            what = "flowchart"
            self["what"].set(what)

        # Remove all the widgets
        frame = self["frame"]
        for slave in frame.grid_slaves():
            slave.grid_forget()

        # and put them back in as needed.
        row = 0
        self["what"].grid(row=row, column=0, sticky=tk.W)
        self["source"].grid(row=row, column=1, sticky=tk.W)
        if source == "local files":
            self["directory"].grid(row=row, column=2, sticky=tk.EW)
            self["get directory"].grid(row=row, column=3, columnspan=2, sticky=tk.E)
        elif source == "previous jobs":
            self["dashboard"].grid(row=row, column=2, sticky=tk.EW)
            self["project"].grid(row=row, column=3, columnspan=2, sticky=tk.E)

            self.dashboards = self.dashboard_handler.dashboards

            # Remove any dashboards that we cannot access
            for dashboard, status in self.dashboard_handler.get_all_status():
                if status != "running":
                    self.dashboards.remove(dashboard)
            if len(self.dashboards) == 0:
                self.current_dashboard = None
                tk.messagebox.showwarning(
                    title="No Dashboards available",
                    message="The are no Dashboards that can be connected to.",
                )
            else:
                dashboard_name = self.config.get("SEAMM open", "dashboard")
                if dashboard_name not in self.dashboards:
                    dashboard_name = self.dashboards[0]
                if (
                    self.current_dashboard is None
                    or self.current_dashboard.name != dashboard_name
                ):
                    self.current_dashboard = self.dashboard_handler.get_dashboard(
                        dashboard_name
                    )
                self._projects = None
                projects = ["any"]
                try:
                    self._projects = self.current_dashboard.projects()
                    projects += [*self._projects.keys()]
                except DashboardConnectionError:
                    projects = []

                self["dashboard"].bind("<<ComboboxSelected>>", "")

                self["dashboard"].config(values=self.dashboards, state="readonly")
                self["dashboard"].set(self.current_dashboard.name)

                self["dashboard"].bind("<<ComboboxSelected>>", self.update_dashboard)

                self["project"].config(values=projects, state="readonly")

                project = self.config.get("SEAMM open", "project")
                if project not in projects:
                    project = projects[0]
                if self.current_project is None or self.current_project != project:
                    self.current_project = project
                self["project"].set(self.current_project)
        row += 1

        if what == "flowchart":
            if "Zenodo" in source:
                self["criteria"].grid(row=row, column=0, columnspan=5, sticky=tk.NSEW)
                frame.rowconfigure(row, weight=1)
                frame.columnconfigure(1, weight=1)
                row += 1
                self["tree"].bind("<<TreeviewOpen>>", "")
                self["search"].grid(row=row, column=0, sticky=tk.W)
                row += 1
                self.clear_tree()
            elif source == "previous jobs":
                self.clear_tree()
                self["tree"].bind("<<TreeviewOpen>>", "")

                # Get the jobs from the Dashboard.
                if self.current_dashboard is not None:
                    if self.current_project == "any":
                        job_list = self.current_dashboard.jobs()
                    else:
                        job_list = self._projects[self.current_project].jobs()
                    self.fill_tree(job_list)
            elif source == "local files":
                self.reset_tree()

            self["tree"].grid(row=row, column=0, columnspan=4, sticky=tk.NSEW)
            self["tree ysb"].grid(row=row, column=4, sticky=tk.NS + tk.W)
            frame.rowconfigure(row, weight=1)
            row += 1

            self["tree xsb"].grid(row=row, column=0, columnspan=4, sticky=tk.EW)
            row += 1

            self["detail"].grid(row=row, column=0, columnspan=5, sticky=tk.NSEW)
            row += 1
        frame.columnconfigure(2, weight=1)

    def reset_tree(self, event=None):
        """Reset the file tree to start with the given directory."""
        directory = self["directory"].get()
        path = Path(directory).expanduser().resolve()

        # Fixes issue #131
        if not path.exists():
            tk.messagebox.showwarning(
                title="Invalid directory",
                message=f"The directory given '{path}' does not exist!",
            )
            return
        elif not path.is_dir():
            if path.is_file():
                path = path.parent
                self["directory"].set(str(path))
            else:
                tk.messagebox.showwarning(
                    title="Not a directory",
                    message=f"The path '{path}' exists, but is not a directory!",
                )
                return
        self.clear_tree()
        node = self.insert_node("", path, path, open=True)
        self["tree"].focus(node)
        self.open_node()
        self["tree"].bind("<<TreeviewOpen>>", self.open_node)

    def open(self):
        """Present a dialog for opening."""
        # Reread the configuration file in case it has changed
        self.config.re_read()

        # Create the dialog if it doesn't exist
        self.create_dialog()

        self.reset_dialog()

        # And put it on-screen, the first time centered.
        result = self["dialog"].activate(geometry="centerscreenfirst")

        if result == "Open":
            what = self["what"].get()
            source = self["source"].get()

            if what == "flowchart":
                tree = self["tree"]
                selected = tree.selection()
                if source != self.config.get("SEAMM open", "source"):
                    self.config.set("SEAMM open", "source", source)
                if source == "Zenodo" or source == "Zenodo sandbox":
                    if len(selected) > 0:
                        record = self._data[selected[0]]["record"]
                        data = record.get_file("flowchart.flow")
                        return {"data": data, "source": source}
                elif "previous jobs" == source:
                    if len(selected) > 0:
                        self.config.set(
                            "SEAMM open", "dashboard", self.current_dashboard.name
                        )
                        self.config.set("SEAMM open", "project", self.current_project)
                        job = self._data[selected[0]]
                        job_id = job["id"]
                        job = self.current_dashboard.job(job_id)
                        data = job.get_file("flowchart.flow")
                        return {"data": data, "source": source, "job_id": job_id}
                elif "local files" in source:
                    if len(selected) > 0:
                        path = self._data[selected[0]]
                        directory = str(path.parent)
                        directories = json.loads(
                            self.config.get("SEAMM open", "directory")
                        )
                        if directory != directories[0]:
                            if directory in directories:
                                directories.remove(directory)
                            directories.insert(0, directory)
                            self.config.set(
                                "SEAMM open", "directory", json.dumps(directories)
                            )
                            w = self["directory"]
                            w.configure(values=directories)
                            w.set(directory)
                        if path.suffix == ".flow":
                            data = path.read_text()
                            return {"data": data, "source": source, "path": str(path)}
                else:
                    raise RuntimeError(f"cannot handle source '{source}'")
            else:
                raise RuntimeError(f"cannot handle opening '{what}'")
        return None

    def search_cb(self):
        """Handle the search."""
        what = self["what"].get()
        source = self["source"].get()

        if what == "flowchart":
            if source == "Zenodo":
                self.search_zenodo_for_flowcharts()
            elif source == "Zenodo sandbox":
                self.search_zenodo_for_flowcharts(sandbox=True)
            else:
                raise RuntimeError(f"Can't handle source '{source}'")
        else:
            raise RuntimeError(f"Can't handle searching for '{what}'")

    def search_zenodo_for_flowcharts(self, sandbox=False):
        """Search for flowcharts in Zenodo.

        Parameters
        ----------
        sandbox : bool = False
            If true, search the Zenodo sandbox.
        """
        self.zenodo = seamm_util.Zenodo(use_sandbox=sandbox)

        criteria = self["criteria"].get()

        query = ""
        after_parenthesis = False
        for i, search_criteria in enumerate(criteria):
            inclusion, field, operator, value, value2 = search_criteria

            logger.debug(f"{inclusion=} {field=} {operator=} {value=} {value2=}")

            zf = zenodo_fields[field]["field"]

            if i == 0:
                # query += " AND ("
                after_parenthesis = True

            pre = ""
            if inclusion == "and":
                if not after_parenthesis:
                    query += " AND"
                continue
            elif inclusion == "or":
                if not after_parenthesis:
                    query += " OR"
                continue
            elif inclusion == "(":
                query += " ("
                after_parenthesis = True
                continue
            elif inclusion == ")":
                query += " )"
                after_parenthesis = False
                continue
            elif inclusion == "must have":
                pre = "+"
            elif inclusion == "must not have":
                pre = "-"
            elif inclusion == "may have":
                pre = ""
            elif inclusion == "ignore":
                continue
            else:
                raise RuntimeError(f"{inclusion=}")

            after_parenthesis = False

            if " " in value:
                value = f'"{value}"'

            if field == "date":
                try:
                    date = dateutil.parser.parse(value).isoformat()
                except Exception:
                    tk.messagebox.showwarning(
                        title="Invalid date format",
                        message=(
                            f"The date given '{value}' can't be handled. Try formats "
                            "like '2022-04-23' or '4/23/2022' or '23 Apr 2022'"
                        ),
                    )
                else:
                    if operator == "after":
                        query += (
                            f' (+keywords:"seamm-flowchart" AND {pre}{zf}[{date} TO *]'
                        )
                    elif operator == "before":
                        query += (
                            f' (+keywords:"seamm-flowchart" AND {pre}{zf}[* TO {date}]'
                        )
                    elif operator == "on":
                        query += f' (+keywords:"seamm-flowchart" AND {pre}{zf}{date}'
                    elif operator == "between":
                        try:
                            date2 = dateutil.parser.parse(value2).isoformat()
                        except Exception:
                            tk.messagebox.showwarning(
                                title="Invalid date format",
                                message=(
                                    f"The date given '{value}' can't be handled. Try "
                                    "formats like '2022-04-23' or '4/23/2022' or "
                                    "'23 Apr 2022'"
                                ),
                            )
                        else:
                            query += (
                                f' (+keywords:"seamm-flowchart" AND {pre}{zf}[{date} to'
                                f" {date2}])"
                            )
                    else:
                        raise RuntimeError(f"Don't recognize operator '{operator}'")
            else:
                if operator == "is":
                    query += f' (+keywords:"seamm-flowchart" AND {pre}{zf}/{value}/)'
                elif operator == "is like":
                    query += f' (+keywords:"seamm-flowchart" AND {pre}{zf}{value}~)'
                elif operator == "contains":
                    query += f' (+keywords:"seamm-flowchart" AND {pre}{zf}{value})'
                elif operator == "contains like":
                    query += f' (+keywords:"seamm-flowchart" AND {pre}{zf}{value}~)'
                elif operator == "is not":
                    query += f' (+keywords:"seamm-flowchart" AND -{zf}/{value}/)'
                elif operator == "is not like":
                    query += f' (+keywords:"seamm-flowchart" AND -{zf}{value}~)'
                elif operator == "does not contain":
                    query += f' (+keywords:"seamm-flowchart" AND -{zf}{value})'
                elif operator == "does not contain like":
                    query += f' (+keywords:"seamm-flowchart" AND -{zf}{value}~)'
                else:
                    raise RuntimeError(f"Don't recognize operator '{operator}'")

        if len(query) == 0:
            query = '+keywords:"seamm-flowchart"'

        logger.debug(f"query = {query}")

        try:
            n_hits, records = self.zenodo.search(query=query, all_versions=True)
        except RuntimeError as e:
            tk.messagebox.showwarning(title="Invalid Zenodo query", message=str(e))
            return

        # Find all related records using conceptrecid
        concepts = {}
        logger.debug(f"Records {n_hits=}")
        for record in records:
            concept_id = record["conceptrecid"]
            if concept_id not in concepts:
                concepts[concept_id] = {}
            data = concepts[concept_id]
            version = record["metadata"]["relations"]["version"][0]["index"]
            data[version] = record

        logger.debug("\n" + pprint.pformat(concepts))
        logger.debug("--------\n\n")

        # Remove any current data
        self.clear_tree()
        self._data = {}
        tree = self["tree"]
        for concept_id, data in concepts.items():
            first = True
            count = len(data)
            for version in sorted(data.keys(), reverse=True):
                record = data[version]
                if first:
                    first = False
                    iid = tree.insert("", "end", text=record.title)
                    self._data[iid] = {
                        "count": count,
                        "record": record,
                    }
                if len(data) > 1:
                    text = f"Version {version + 1}: {record.title}"
                    jid = tree.insert(iid, "end", text=text)
                    self._data[jid] = {
                        "count": count,
                        "record": record,
                    }

    def select_record(self, event):
        """The user clicked on the tree-view ... handle the selected record."""
        tree = self["tree"]
        selected = tree.selection()
        if len(selected) == 0:
            return

        source = self["source"].get()

        if "Zenodo" in source:
            tmp = self._data[selected[0]]
            count = tmp["count"]
            data = tmp["record"]["metadata"]
            logger.debug("\n\nZenodo source, data =")
            logger.debug("\n" + pprint.pformat(data))
            logger.debug("-------")
            self["title"].configure(text=data["title"])
            self["description"].delete("1.0", "end")
            self["description"].insert("1.0", data["description"])
            info = data["relations"]["version"][0]
            if "publication_data" in data:
                date = data["publication_data"]
                version = f"Version {info['index'] + 1} of {count} -- {date}"
            else:
                version = f"Version {info['index'] + 1} of {count}"
            self["version"].configure(text=version)
        elif "previous jobs" == source:
            job = self._data[selected[0]]
            self["title"].configure(text=job["title"])
            self["description"].delete("1.0", "end")
            self["description"].insert("1.0", job["description"])
        elif source == "local files":
            self["title"].configure(text="")
            self["description"].delete("1.0", "end")
            self["version"].configure(text="")
            path = self._data[selected[0]]
            if path.suffix == ".flow":
                capture = False
                lines = []
                with path.open() as fd:
                    for line in fd:
                        line = line.strip()
                        if line[0] == "#":
                            if line == "#metadata":
                                capture = True
                            elif capture:
                                break
                        elif capture:
                            lines.append(line)
                if len(lines) > 0:
                    mtime = path.stat().st_mtime
                    date = datetime.datetime.fromtimestamp(mtime).isoformat(" ")
                    data = json.loads("\n".join(lines))
                    if "title" in data:
                        self["title"].configure(text=data["title"])
                    if "description" in data:
                        self["description"].insert("1.0", data["description"])
                    if "version" in data:
                        line = f"Version {data['version']} -- {date}"
                    else:
                        line = f"Version <none> -- {date}"
                    self["version"].configure(text=line)
        else:
            raise RuntimeError(f"Can't handle source '{source}'")

    def update_dashboard(self, event=None):
        """The dashboard has been changed!"""

        dashboard = self["dashboard"].get()
        project = self["project"].get()

        if dashboard != self.current_dashboard.name:
            projects = ["any"]
            try:
                dashboard = self.dashboard_handler.get_dashboard(dashboard)
                self._projects = self.current_dashboard.projects()
                projects += [*self._projects.keys()]
            except DashboardConnectionError:
                self.dashboards.remove(dashboard)
                self["dashboard"].config(values=self.dashboards, state="readonly")
                self["dashboard"].set(self.current_dashboard.name)
                tk.messagebox.showwarning(
                    title="Cannot access Dashboard",
                    message=f"Cannot access Dashboard {dashboard}. Resetting!",
                )
            else:
                self.current_dashboard = dashboard
                self["project"].config(values=projects, state="readonly")

                if self.current_project is None or project not in projects:
                    self.current_project = projects[0]
                self["project"].set(self.current_project)
        elif project != self.current_project:
            self.current_project = project

        self.clear_tree()
        # Get the jobs from the Dashboard.
        if self.current_project == "any":
            job_list = self.current_dashboard.jobs()
        else:
            job_list = self._projects[self.current_project].jobs()
        self.fill_tree(job_list)

    def zenodo_callback(self, widget, criterion, event, what):
        if criterion is not None:
            inclusion, field, operator, value, value2 = criterion.get()

            if what == "field":
                operators = zenodo_fields[field]["operators"]
                w = criterion.operator
                w.configure(values=operators)
                if operator in operators:
                    w.set(operator)
                else:
                    w.set(operators[0])
